from copy import copy
from urllib.parse import unquote
from http.cookies import SimpleCookie
from email.parser import BytesParser as BP
try:
	from orjson import loads  # type: ignore
except ModuleNotFoundError:
	try:
		from ujson import loads  # type: ignore
	except ModuleNotFoundError:
		from json import loads


class Request:
	__slots__ = (
		"_scope", "_event", "_scheme", "_path",
		"_query_string", "_headers", "_ip", "_data",
		"_type", "_content_length"
	)

	def __init__(self, scope, event, content_type):
		self._scope = scope
		self._event = event
		self._query_string = None
		self._type = content_type
		self._headers = None
		self._data = None

	@staticmethod
	def _parse(data: bytes) -> dict:
		if not data:
			return {}
		parsed: dict = {}
		for item in unquote(data.decode()).split("&"):
			pair = item.split("=") + [""]
			if pair[0] in parsed:
				if type(parsed[pair[0]]) == list:
					parsed[pair[0]].append(pair[1])
				else:
					parsed[pair[0]] = [parsed[pair[0]], pair[1]]
			else:
				parsed[pair[0]] = pair[1]
		return parsed


	@property
	def scope(self):
		return self._scope

	@property
	def event(self):
		return self._event

	@property
	def path(self):
		return self._scope["path"]

	@property
	def query_params(self):
		if self._query_string is None:
			self._query_string = self._parse(self._scope["query_string"])
		return copy(self._query_string)

	@property
	def headers(self):
		if self._headers is None:
			self._headers = {}
			for header, value in self._scope["headers"]:
				self._headers[header.decode()] = value.decode()
		return copy(self._headers)

	@property
	def data(self):
		if self._data is None:
			if self._type[:16] == "application/json":
				try:
					self._data = loads(self._event.get("bytes") or self._event.get("text").encode())
				except Exception:
					raise ParseError
			else:
				self._data = self._event.get("text") or self._event["bytes"].decode()
		return self._data


class HTTPRequest(Request):
	__slots__ = (
		"_scope", "_event", "_scheme", "_path",
		"_query_string", "_headers", "_ip", "_data",
		"_version", "_method", "_cookies"
	)

	def __init__(self, scope, event):
		super().__init__(scope, event, "")
		self._cookies = None

	@property
	def method(self):
		return self._scope["method"]

	@property
	def cookies(self):
		if self._cookies is None:
			cookie: SimpleCookie = SimpleCookie()
			for header, value in self._scope["headers"]:
				if header.lower() == b"cookie":
					cookie.load(value.decode())
			self._cookies = {k: v for k, v in cookie.items()}
		return copy(self._cookies)

	@property
	def headers(self):
		if self._headers is None:
			self._headers = {}
			for header, value in self._scope["headers"]:
				if header.lower() != b"cookie":
					self._headers[header.decode()] = value.decode()
		return copy(self._headers)

	@property
	def data(self):
		if self._data is None:
			try:
				if not int(self.headers["content-length"]):
					self._data = {}
				elif self._headers["content-type"][:16] == "application/json":
					self._data = loads(self._event["body"])
				elif self._headers["content-type"] == "application/x-www-form-urlencoded":
					self._data = self._parse(self._event["body"])
				elif self._headers["content-type"][:19] == "multipart/form-data":
					m = BP().parsebytes(
						b"Content-Type: " + self._headers["content-type"].encode() +
						b"\n" + self._event["body"]
					)
					self._data = {
						p.get_param("name", header="content-disposition"): p.get_payload(decode=True) for p in m.get_payload()
					}
				else:
					self._data = self._event["body"]
			except Exception:
				raise ParseError
		return self._data


class InputParseError(Exception):
	pass



from .request import HTTPRequest, Request, InputParseError
from .response import HTTPResponse, Response, HTTPAbort
from .route import HTTPRoute, WSRoute


class ASGI:
	__slots__ = ()
	_content_type = ""
	_charset = ""
	_allow = None
	triggers: dict = {}

	def __init__(self, content_type: str = "application/json", charset: str = "UTF-8", allow=""):
		ASGI._content_type = content_type
		ASGI._charset = charset
		ASGI._allow = allow

	@staticmethod
	async def __call__(scope, receive, send):
		event = await receive()
		if scope["type"] == "http":
			while event["more_body"]:
				more = await receive()
				event["body"] += more["body"]
				event["more_body"] = more["more_body"]
			await ASGI._HTTP(scope, event, send)
		elif event["type"] == "websocket.connect":
			await ASGI._WS(scope, send, receive)

	@staticmethod
	async def _HTTP(scope, event, send):
		method, url_args, not_allowed, ct = HTTPRoute.get(scope["path"], scope["method"])

		request = HTTPRequest(scope, event)
		response = HTTPResponse(
			send,
			content_type=ct or ASGI._content_type,
			charset=ASGI._charset,
			allow=ASGI._allow
		)
		try:
			if not_allowed:
				await response.abort(status=405)
			elif not method:
				await response.abort(status=404)
			try:
				body = await method(request, response, *url_args)
			except InputParseError:
				await response.abort(status=400)
			if not response.processed:
				await response.process(body)
		except HTTPAbort:
			pass

	@staticmethod
	async def _WS(scope, send, receive):
		trigger, url_args, ct = WSRoute.get(scope["path"])
		if trigger:
			await send({"type": "websocket.accept"})
			while True:
				event = await receive()
				request = Request(scope, event, content_type=ASGI._content_type)
				response = Response(
					send,
					content_type=ct or ASGI._content_type,
					charset=ASGI._charset
				)
				try:
					await trigger(request, response, url_args)
				except InputParseError:
					response.send({"status": False, "error": "P001", "message": "Data JSON parse error"})
				if event["type"] == "websocket.disconnect":
					break
		else:
			await send({"type": "websocket.close"})



from re import Pattern


class Route:
	__slots__ = ("_instance")

	def __init__(self, *args, type: str = "http", **kwargs):
		if type.lower() == "http":
			self._instance = HTTPRoute(*args, **kwargs)
		elif type.lower() == "websocket":
			self._instance = WSRoute(*args, **kwargs)

	def __call__(self, call, *args):
		return self._instance(call, *args)


class WSRoute:
	__slots__ = ("_route", "_content_type")
	__routes: dict = {}
	__regex_routes: list = []

	def __init__(self, route, content_type=None):
		self._content_type = content_type
		if type(route) == str:
			if route[0] != "/":
				raise Exception("Route must start with '/'! (%s)" % route)
			self._route = route if route[-1] == ("/") else route + "/"
		elif type(route) == Pattern:
			self._route = route
		else:
			raise Exception("Route must be str or re.Pattern type! (%s, %s)" % (route, str(type(route))))

	def __call__(self, fce, *args):
		if type(self._route) == Pattern:
			self.__regex_routes.append((self._route, fce, self._content_type))
		else:
			if self._route not in self.__routes:
				self.__routes[self._route] = (fce, self._content_type)
			else:
				raise Exception("Route alreay set! (%s)" % self._route)
		return fce

	@staticmethod
	def get(path: str) -> tuple:
		route = WSRoute.__routes.get(path if path[-1] == "/" else path + "/", {})
		if route:
			return route[0], None, route[1]
		else:
			for item in WSRoute.__regex_routes:
				if match := item[0].match(path):
					return item[1], match.groups(), item[2]
		return None, None


class HTTPRoute(WSRoute):
	__slots__ = ("_route", "__methods", "_content_type")
	__routes: dict = {}
	__regex_routes: list = []

	def __init__(self, route, methods: list = [], content_type=None):
		super().__init__(route, content_type)
		self.__methods = methods or ["GET"]

	def __call__(self, fce, *args):
		if type(self._route) == Pattern:
			self.__regex_routes.append((self._route, fce, self.__methods, self._content_type))
		else:
			for method in self.__methods:
				if self._route not in self.__routes:
					self.__routes[self._route] = {}
				elif method in self.__routes[self._route]:
					raise Exception("Route alreay set! (%s)" % self.__routes.get(method, self._route))
				self.__routes[self._route][method] = (fce, self._content_type)
		return fce

	@staticmethod
	async def _options(request, response):
		path = request.path
		methods = HTTPRoute.__routes.get(path if path[-1] == "/" else path + "/", {}).keys()
		if not methods:
			for item in HTTPRoute.__regex_routes:
				if item[0].match(path):
					methods = item[2]
					break
		response.headers.append((b"Access-Control-Allow-Headers", b"*"))
		response.headers.append((b"Access-Control-Allow-Methods", ",".join(methods).encode()))
		response.headers.append((b"Vary", b"Access-Control-Request-Headers"))
		return b""

	@staticmethod
	def get(path: str, method: str) -> tuple:  # type: ignore[override]
		if method == "OPTIONS":
			return HTTPRoute._options, (), False, "text/plain"
		route = HTTPRoute.__routes.get(path if path[-1] == "/" else path + "/", {})
		if result := route.get(method, False):
			return result[0], (), None, result[1]
		else:
			for item in HTTPRoute.__regex_routes:
				if match := item[0].match(path):
					route = True
					if method in item[2]:
						return item[1], match.groups(), None, item[3]
		if route and not result:
			return None, (), True, None
		return None, (), False, None



#!/usr/bin/python
import sys
import uvicorn
# from gunicorn.app.wsgiapp import run
from argparse import ArgumentParser


def server():
	parser = ArgumentParser(
		description="Simple development server for WeefPy framework"
	)
	parser.add_argument('--daemon', action="store_true", help="Run server as daemon")
	parser.add_argument('--config', nargs=1, default=[""], help="Path to gunicorn config file")
	parser.add_argument('--pythonpath', nargs=1, default=["."], help="Provided pythonpath")
	parser.add_argument('--host', nargs=1, default=[None], help="Server IP address")
	parser.add_argument('--port', nargs=1, default=[None], help="Server port")
	parser.add_argument(
		'--app', nargs=1, default=[None],
		help="Callable application in format file:app - where app is callable object"
	)
	args = parser.parse_args(sys.argv[1:])
	config = args.config[0]
	pythonpath = args.pythonpath[0]
	host = args.host[0] or "127.0.0.1"
	port = int(args.port[0] or "8000")
	app = args.app[0]

	# options = {
	# 	'bind': '%s:%s' % (host, port),
	# 	'workers': "4",
	# 	'pythonpath': pythonpath,
	# 	'worker-class': "uvicorn.workers.UvicornWorker"
	# }
	# if config:
	# 	options["config"] = config

	# sys.argv = [sys.argv[0]]
	# for key in options:
	# 	sys.argv.append("--%s" % key)
	# 	sys.argv.append(str(options[key]))
	# sys.argv.append(app)

	options = {}
	if config:
		options["env_file"] = config

	if args.daemon:
		import daemon
		with daemon.DaemonContext():
			# sys.exit(run())
			uvicorn.run(
				app,
				host=host,
				port=port,
				reload=True,
				app_dir=pythonpath,
				workers=4,
				log_level="info",
				timeout_keep_alive=3600,
				**options
			)
	else:
		# sys.exit(run())
		uvicorn.run(
			app,
			host=host,
			port=port,
			reload=True,
			app_dir=pythonpath,
			workers=4,
			log_level="info",
			timeout_keep_alive=3600,
			**options
		)




from datetime import datetime
try:
	from orjson import dumps  # type: ignore
except ModuleNotFoundError:
	try:
		from ujson import dumps as jdumps  # type: ignore
	except ModuleNotFoundError:
		from json import dumps as jdumps

	def dumps(*args, **kwargs):  # noqa
		return jdumps(*args, **kwargs).encode()


class Response:
	__slots__ = ("_send", "_type", "_redirect", "charset")

	def __init__(self, send, content_type="application/json", charset="UTF-8", allow=None):
		self._redirect = False
		self._send = send
		self._type = content_type
		self.charset = charset

	async def send(self, data):
		if self._type == "application/json" and type(data) in (list, dict):
			data = dumps(data).decode()
		if type(data) == bytes:
			await self._send({"type": "websocket.send", "body": data})
		else:
			await self._send({"type": "websocket.send", "text": data})

	@property
	def content_type(self) -> str:
		return self._type


class HTTPResponse(Response):
	__slots__ = ["_processed", "_headers", "_allow"]

	def __init__(self, send, content_type, charset, allow):
		self._processed: bool = False
		self._headers: list[tuple] = [(b"Access-Control-Allow-Origin", allow.encode())]
		self._allow = allow
		super().__init__(send, content_type, charset)

	async def process(self, data, status=200):
		if self._redirect:
			return
		await self.start(status)
		if self._type == "application/json" and type(data) in (list, dict):
			body = dumps(data)
		elif type(data) == bytes:
			body = data
		elif type(data) == str:
			body = bytes(data, self.charset)
		else:
			body = b"" if data is None else bytes(str(data), self.charset)
		await self._send({"type": "http.response.body", "body": body})
		self._processed = True

	async def abort(self, status=400, data=b""):
		await self.process(data, status=status)
		raise HTTPAbort()

	async def start(self, status=200):
		if self._type[:4] == "text" or self._type[:11] == "application":
			self.add_header('Content-Type', self._type + ";charset=" + self.charset)
		else:
			self.add_header('Content-Type', self._type)
		await self._send({
			"type": "http.response.start",
			"status": status,
			"headers": self._headers
		})

	async def send(self, data: bytes):  # type: ignore
		await self._send({"type": "http.response.body", "body": data})

	async def redirect(self, location: str, status: int = 301) -> None:
		self._redirect = True
		self.add_header('Location', location)
		await self.start(status=status)

	def set_cookie(
			self, name: str, value: str, expires: datetime = None, maxAge="", **kwargs) -> None:
		cookie = name + "=" + value
		if expires:
			cookie += "; Expires=" + expires.strftime('%a, %d %b %Y %H:%M:%S %Z')
		elif maxAge:
			cookie += "; Max-Age=" + str(maxAge)
		for arg in kwargs:  # Domain=None,Path=None,Secure=False,HttpOnly=False
			cookie += "; " + arg[0].upper() + arg[1:] + "=" + kwargs[arg] if kwargs[arg] else "; " + arg[0].upper() + arg[1:]
		self.add_header("Set-Cookie", cookie)

	def add_header(self, name: str, value: str) -> None:
		self._headers.append((name.encode(), value.encode()))

	@property
	def processed(self):
		return self._processed

	@property
	def headers(self) -> list:
		return self._headers

	@property
	def content_type(self) -> str:
		return self._type


class HTTPAbort(Exception):
	pass



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="weepy",
	version="1.0.11",
	scripts=["tools/dev_server.py"],
	entry_points={
		'console_scripts': ['weepy=dev_server:server'],
	},
	author="Patrik Katrenak",
	author_email="patrik@katryapps.com",
	description="Tiny Python ASGI framework",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://gitlab.com/katry/weepy",
	packages=["weepy"],
	install_requires=[],
	extras_require={
		"orjson": ["orjson"],
		"ujson": ["ujson"],
		"dev": ["daemon", "uvicorn[standard]", "orjson", "python-dotenv", "requests", "pytest"]
	},
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
		"Operating System :: OS Independent",
	],
	platforms=["any"],
	python_requires=">=3.9",
)
